package hu.bme.mit.mobilegen.datamodel.ui.builder;

import static com.google.common.collect.Maps.uniqueIndex;
import static org.eclipse.xtext.ui.util.ResourceUtil.getContainer;
import hu.bme.mit.mobilegen.datamodel.datamodel.Content;
import hu.bme.mit.mobilegen.datamodel.datamodel.Datamodel;
import hu.bme.mit.mobilegen.datamodel.datamodel.DatamodelFactory;
import hu.bme.mit.mobilegen.datamodel.datamodel.Element;
import hu.bme.mit.mobilegen.datamodel.datamodel.impl.DatamodelFactoryImpl;
import hu.bme.mit.mobilegen.datamodel.generator.GeneratorConstants;
import hu.bme.mit.mobilegen.datamodel.generator.GeneratorData;
import hu.bme.mit.mobilegen.datamodel.generator.IDatamodelGenerator;
import hu.bme.mit.mobilegen.datamodel.helper.Utility;
import hu.bme.mit.mobilegen.datamodel.helper.WrappedModel;
import hu.bme.mit.mobilegen.datamodel.logger.Log;
import hu.bme.mit.mobilegen.datamodel.preprocessor.ModelValidator;
import hu.bme.mit.mobilegen.datamodel.preprocessor.PreProcessor;

import java.io.File;
import java.io.IOException;
import java.util.Collections;
import java.util.Date;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.emf.common.notify.Notifier;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.xtext.EcoreUtil2;
import org.eclipse.xtext.builder.EclipseResourceFileSystemAccess2;
import org.eclipse.xtext.generator.OutputConfiguration;
import org.eclipse.xtext.ui.resource.IResourceSetProvider;

import com.google.common.base.Function;
import com.google.common.collect.Iterators;
import com.google.inject.Inject;
import com.google.inject.Provider;

public class DatamodelBuilder {

	@Inject
	private Provider<EclipseResourceFileSystemAccess2> fileSystemAccessProvider;

	@Inject
	private IDatamodelGenerator generator;

	@Inject
	private IResourceSetProvider resourceSetProvider;

	@Inject
	private DatamodelOutputConfigurations outputConfigurations;

	private IProject developerProject;
	private IProject targetProject;

	private Date startTime;
	private Date endTime;

	private IResource[] srcResources;
	private ResourceSet rs;

	private EclipseResourceFileSystemAccess2 access;

	private Map<String, OutputConfiguration> outputConfig;

	public void generateDataModel(IProject project) {
		startTime = new Date();
		Log.raw(Utility.formatTime(startTime) + " - STARTING DATAMODEL GENERATION");

		developerProject = project;

		rs = resourceSetProvider.get(developerProject);

		if (!checkDeveloperProject(developerProject)) {
			return;
		}

		if (!parseConfiguration()) {
			return;
		}

		if (!searchModelFiles()) {
			return;
		}

		if (!configureTargetProject()) {
			return;
		}

		if (!processModel()) {
			return;
		}

		endTime = new Date();

		Log.separator();
		Log.raw(Utility.formatTime(endTime) + " - GENERATION PROCESS ENDED SUCCESSFULLY");
	}
	
	public void exportDataModelToXMI(IProject project) {
		startTime = new Date();
		Log.raw(Utility.formatTime(startTime) + " - STARTING DATAMODEL EXPORTING TO XMI");

		developerProject = project;

		rs = resourceSetProvider.get(developerProject);

		if (!checkDeveloperProject(developerProject)) {
			return;
		}

		if (!parseConfiguration()) {
			return;
		}

		if (!searchModelFiles()) {
			return;
		}

		if (!exportModelToXMI()) {
			return;
		}

		endTime = new Date();

		Log.separator();
		Log.raw(Utility.formatTime(endTime) + " - EXPORTING PROCESS ENDED SUCCESSFULLY");
	}

	public void generateDatamodelFromXMI(IFile file) {
		startTime = new Date();
		Log.raw(Utility.formatTime(startTime) + " - STARTING DATAMODEL GENERATION");

		developerProject = file.getProject();

		rs = new ResourceSetImpl();

		if (!loadXMIResource(file)) {
			return;
		}

		if (!parseConfiguration()) {
			return;
		}

		if (!configureTargetProject()) {
			return;
		}

		if (!processModel()) {
			return;
		}

		endTime = new Date();

		Log.separator();
		Log.raw(Utility.formatTime(endTime) + " - GENERATION PROCESS ENDED SUCCESSFULLY");
	}

	private boolean checkDeveloperProject(IProject developerProject) {
		Log.header("STARTING TO CHECK DATAMODEL PROJECT");

		Log.info("Checking developer project");
		IFolder src = developerProject.getFolder(BuilderConstants.SRC_FOLDER);

		if (!src.exists()) {
			Log.error("src folder not found in Modeling project - " + developerProject.getFullPath().toString());
			Log.terminated();
			return false;
		}

		try {
			srcResources = src.members();
		} catch (CoreException e) {
			Log.error("Datamodel files not found in project - " + developerProject.getFullPath().toString());
			Log.terminated();
			return false;
		}

		Log.info("Checking project ended successfully");
		Log.separator();

		return true;
	}

	private boolean parseConfiguration() {
		if (!ConfigurationParser.getConfigurations(developerProject)) {
			Log.error("Error in parsing config file");
			Log.terminated();
			return false;
		}

		Log.info("Parsing configuration file ended successfully");
		Log.separator();
		return true;
	}

	private boolean searchModelFiles() {
		Log.header("STARTING SEARCH FOR DATAMODEL FILES");

		int resources = 0;
		for (IResource r : srcResources) {
			if (r instanceof IFile) {
				if (((IFile) r).getFileExtension().equals(FileExtensions.DATAMODEL_EXTENSION)) {
					Log.info("Found datamodel resource - " + r.getFullPath().toString());
					IFile file = (IFile) r;
					rs.getResource(URI.createURI(file.getFullPath().toString()), true);
					resources++;
				}
			}
		}

		if (resources == 0) {
			Log.error("No datamodel files found in project - " + developerProject.getFullPath().toString());
			Log.terminated();
			return false;
		}

		Log.info("Datamodel files collected");
		Log.separator();

		return true;
	}

	private boolean configureTargetProject() {
		Log.header("SARTING TO CONFIGURE TARGET PROJECT - " + GeneratorData.getProjectName());

		access = fileSystemAccessProvider.get();
		targetProject = ResourcesPlugin.getWorkspace().getRoot().getProject(GeneratorData.getProjectName());

		if (!targetProject.exists()) {
			Log.error(GeneratorData.getProjectName() + " project does not exists in this Workspace");
			Log.terminated();
			return false;
		}

		Log.info("Target project is: " + GeneratorData.getProjectName());

		GeneratorData.setProject(targetProject);
		access.setProject(targetProject);

		outputConfig = getOutputConfigurations(targetProject);

		access.setOutputConfigurations(outputConfig);
		access.setMonitor(new NullProgressMonitor());

		Log.info("Setting up configurations succedded");
		Log.separator();

		return true;
	}

	private boolean exportModelToXMI() {
		Log.info("Exporting datamodel to xmi file");

		ResourceSet resourceSet = new ResourceSetImpl();
		EcoreUtil2.clone(resourceSet, rs);

		URI fileURI = URI.createFileURI(new File(developerProject.getLocation() + "\\" + developerProject.getName() + "_datamodel.xmi").getAbsolutePath());
		Resource resource = resourceSet.createResource(fileURI);

		TreeIterator<Notifier> content = resourceSet.getAllContents();
		Iterator<Element> elements = Iterators.<Element> filter(content, Element.class);

		DatamodelFactory factory = new DatamodelFactoryImpl();
		Datamodel cm = factory.createDatamodel();
		Content c = factory.createContent();
		cm.setContent(c);

		while (elements.hasNext()) {
			Element e = elements.next();
			c.getElement().add(e);
		}

		resource.getContents().add(cm);

		try {
			resource.save(Collections.EMPTY_MAP);
			developerProject.refreshLocal(IResource.DEPTH_INFINITE, null);
			Log.info("XMI File generated - File location: " + fileURI.toFileString());
		} catch (IOException | CoreException e) {
			Log.warning("Exporting to XMI Failed");
			return true;
		}

		return true;
	}

	private boolean processModel() {
		try {

			Log.header("STARTING CODE GENERATION PROCESS");

			WrappedModel wm = PreProcessor.preProcessModel(rs);

			if (!ModelValidator.checkModel(wm)) {
				Log.terminated();
				return false;
			}

			if (!cleanOutput(outputConfig, targetProject)) {
				Log.terminated();
				return false;
			}

			Log.separator();
			Log.header("STARTING GENERATION");

			generator.generate(wm, access);

			targetProject.refreshLocal(IResource.DEPTH_INFINITE, null);

			return true;
		} catch (RuntimeException | CoreException e) {
			Log.error("Error occured during generation process");
			Log.terminated();
			return false;
		}

	}

	private Map<String, OutputConfiguration> getOutputConfigurations(IProject project) {
		Log.info("Setting up output configurations");

		Set<OutputConfiguration> configurations = outputConfigurations.getOutputConfigurations(project);
		return uniqueIndex(configurations, new Function<OutputConfiguration, String>() {
			public String apply(OutputConfiguration from) {
				return from.getName();
			}
		});
	}

	private boolean cleanOutput(Map<String, OutputConfiguration> outputConfigurations, IProject project) {
		Log.info("Start cleaning target project - " + GeneratorConstants.SRC_GEN);

		if (!cleanOutputDirectory(project, outputConfigurations.get(GeneratorConstants.SRC_GEN_ENTITIES).getOutputDirectory())) {
			return false;
		}
		if (!cleanOutputDirectory(project, outputConfigurations.get(GeneratorConstants.SRC_GEN_SER).getOutputDirectory())) {
			return false;
		}
		if (!cleanOutputDirectory(project, outputConfigurations.get(GeneratorConstants.SRC_GEN_PREFS).getOutputDirectory())) {
			return false;
		}
		if (!cleanOutputDirectory(project, outputConfigurations.get(GeneratorConstants.SRC_GEN_DB).getOutputDirectory())) {
			return false;
		}

		Log.info("Cleaning was successfull");
		return true;
	}

	private boolean cleanOutputDirectory(IProject project, String path) {
		Log.info("Cleaning - " + path);

		IContainer container = getContainer(project, path);
		if (container.exists()) {
			try {
				for (IResource r : container.members()) {
					r.delete(IResource.KEEP_HISTORY, null);
				}

				return true;
			} catch (CoreException e) {
				Log.error("Error occured while cleaning output - " + path);
				return false;
			}
		}

		return true;
	}

	private boolean loadXMIResource(IFile file) {
		Log.header("STARTING TO LOAD DATAMODEL FROM XMI FILE");

		Log.info("Loading resource - " + file.getFullPath().toString());
		Resource resource = rs.getResource(URI.createURI(file.getFullPath().toString()), true);

		if (resource == null) {
			Log.error("Error occured while loading resource - " + file.getFullPath().toString());
			return false;
		} else {
			Log.info("Resource loaded");
			return true;
		}
	}

}
